from typing import Tuple


def merge_wmi(new_wmis: dict) -> Tuple[set, str]:
    """Helper. Used to update vininfo.dicts.wmi.WMI dictionary
    with entries from another dictionary.

    Existing keys are skipped, new keys are added, result is sorted.

    Returns a two-items tuple:
        0. set of new keys borrowed from `new_wmis`;
        1. source code string to replace the current `WMI` dictionary with.

    :param new_wmis: Mapping from WMI (two or three chars) to manufacturer title.

    """
    import inspect
    import re

    from .dicts import wmi

    wmi_src = re.search('WMI = {([^}]+)}', inspect.getsource(wmi), re.MULTILINE)

    assert wmi_src, 'Unable to parse WMI dict body'

    wmi_src_dict = {}

    for line in wmi_src.group(1).splitlines():
        line = line.strip(', ')

        if not line:
            continue

        key, _, value = line.partition(':')
        wmi_src_dict[key.strip('\'"')] = value.strip()

    wmi_missing = set(new_wmis.keys()).difference(wmi_src_dict.keys())

    for key in wmi_missing:
        wmi_src_dict[key] = f"'{new_wmis[key]}'"

    wmi_dst = ['WMI = {']

    for key, value in sorted(wmi_src_dict.items(), key=lambda item: item[0]):
        wmi_dst.append(f"    '{key}': {value},")

    wmi_dst.append('}')

    return wmi_missing, '\n'.join(wmi_dst)



from ._base import VinDetails, Detail
from ..dicts.bodies import *


class NissanDetails(VinDetails):
    """Nissan VIN details extractor."""

    model = Detail(('vds', 1), {
        'A': ['Armada', 'Titan', 'Maxima'],
        'B': 'Sentra',
        'C': 'Versa (07-11)',
        'D': ['Truck', 'Xterra (00-04)', 'Frontier'],
        'J': 'Maxima',
        'N': 'Xterra (05-11)',
        'R': 'Pathfinder',
        'S': ['240SX', 'Rogue (08-11)'],
        'U': 'Altima',
        'Z': ['300Z', '350Z', 'Murano'],
    })

    body = Detail(('vds', 3), {
        '1': [BODY_SEDAN_4, 'Standard Body Truck'],
        '4': BODY_COUPE_2,
        '5': BODY_SW_5,
        '6': [BODY_CABRI_2, 'Fastback', 'King Cab Truck'],
        '7': 'Crew Cab Truck',
        '8': BODY_SW_8,
    })

    engine = Detail(('vds', 0), {
        'A': ['VG30D', 'VK45DE', 'VQ35DE', 'VK56DE', 'VQ40DE', 'QR25DE'],
        'B': ['KA24DE', 'SR20DE', 'VQ35HR', 'MR18DE', 'QR25DE'],
        'C': ['SR20DE', 'VG30DETT', 'QG18DE'],
        'D': ['KA24DE', 'QG18DE', 'VQ35DE'],
        'E': ['VE30DE', 'GA16DE', 'VG33E'],
        'F': 'KA24E',
        'H': 'VG30E',
        'M': ['KA24DE', 'VG33ER'],
        'N': 'VH45DE',
        'R': 'VG30DE',
        'S': 'KA24E',
        'T': 'VG33E',
    })

    plant = Detail(('vis', 1), {
        'C': 'Smyrna',
        'L': 'Aguas Calientes',
        'M': 'Tochigi',
        'N': 'Canton',
        'T': ['Tochigi', 'Oppama'],
        'W': 'Kyushyu',
    })

    serial = Detail(('vis', slice(2, None)))



from ._base import VinDetails, Detail
from ..dicts.bodies import *


def get_engine(details):

    candidates = {
        'P': {
            '1': 'A18XER140HP',
            'B': 'A14XER100HP',
            'C': 'A14NET140HP',
            'D': 'A16XER115HP',
            'J': 'A16LET180HP',
            'N': 'A20DTH165HP',
            'U': 'A14NEL120HP',
        },
        'W': {
            'N': 'A20DTH160HP',
            'C': 'A14NET140HP',
        },
        'G': {
            'A': 'A16XER115HP',
            'B': 'A16LET180HP',
            'C': 'A18XER140HP',
            'D': 'A20NHT220HP',
            'E': 'A20NHT250HP',
            'F': 'A28NET260HP',
            'G': 'A28NER325HP',
            'M': 'A20DTH160HP',
            'N': 'A20DTR195HP',
            'P': 'A14NFT140HP',
            'X': 'A20NHT250HP',
        },
        'V': {
            '1': 'Y13D70HP',
            '2': 'Z14XEP100HP',
        },
        'S': {
            'C': 'Z14XEP100HP',
            'D': 'A14NEL120HP',
            'E': 'A14NET140HP',
        },
        'J': {
            '8': 'A14NET140HP',
        },
    }

    return candidates.get(details.model.code, {})


class OpelDetails(VinDetails):
    """Opel VIN details extractor."""

    model = Detail(('vds', 0), {
        'F': 'Agila',
        'G': 'Insignia',
        'J': 'Mokka',
        'L': 'Antara',
        'M': 'Movano',
        'P': ['Astra J', 'Zafira C'],
        'R': 'Astra GTC J',
        'S': 'Meriva',
        'V': 'Combo II',
        'W': 'Cascada',
    })

    body = Detail(('vds', 2), {
        '2': BODY_HATCH_3,
        '3': BODY_COUPE_2,
        '5': BODY_SEDAN_4,
        '6': BODY_HATCH_5,
        '7': BODY_CROSS_3,
        '8': BODY_SW_5,
        '9': BODY_MINIVAN_5,
        'B': BODY_MINIBUS,
        'C': BODY_VAN,
        'J': BODY_VAN,
        'X': BODY_SW_3,
    })

    engine = Detail(('vds', 4), get_engine)

    plant = Detail(('vis', 1), {
        '1': 'Russelsheim',
        '2': 'Bochum',
        '3': 'Azambuja',
        '4': 'Zaragoza',
        '5': 'Antwerp',
        '6': 'Eisenach',
        '7': 'Luton',
        '8': 'Ellesmere Port',
        '9': 'Uusikaupunki',
        'A': 'Azambuja',
        'B': 'Bertone / Saint Peterburg',
        'C': 'Yelabuga',
        'D': 'Shin Chuang',
        'E': 'Heuliez',
        'F': 'Togliatti',
        'G': 'Gliwice',
        'H': 'Rayong',
        'L': 'Elizabeth',
        'M': 'Millbrook',
        'N': 'Norwich',
        'P': 'Warsaw',
        'R': 'Rosario',
        'S': 'Szentgotthard',
        'V': 'Luton',
        'X': 'Zaporozhia',
        'Z': 'Izmir',
    })

    serial = Detail(('vis', slice(2, None)))



from vininfo import Vin


def test_opel():

    vin = Vin('W0LPC6DB3CC123456')

    assert '%s' % vin
    assert vin.wmi == 'W0L'
    assert vin.manufacturer == 'Opel/Vauxhall'
    assert vin.vds == 'PC6DB3'
    assert vin.vis == 'CC123456'
    assert vin.years == [2012, 1982]
    assert vin.region_code == 'W'
    assert vin.region == 'Europe'
    assert vin.country_code == 'W0'
    assert vin.country == 'Germany'
    assert '%s' % vin.brand == 'Opel (Opel/Vauxhall)'

    details = vin.details
    assert details.model.code == 'P'
    assert details.model.name == ['Astra J', 'Zafira C']
    assert details.body.code == '6'
    assert details.body.name == 'Hatchback, 5-Door'
    assert details.engine.code == 'B'
    assert details.engine.name == 'A14XER100HP'
    assert details.plant.code == 'C'
    assert details.plant.name == 'Yelabuga'
    assert details.serial.code == '123456'



from typing import Optional, Type

from ..common import Annotatable

if False:  # pragma: nocover
    from ..toolbox import Vin  # noqa


class DetailWrapper:

    __slots__ = ['code', 'name', '_supported']

    def __init__(self, details: 'VinDetails', detail: 'Detail'):
        """
        :param details:
        :param detail:

        """
        vin = details._vin

        source = detail.source

        code = ''

        if source:
            attr_name, attr_idx = source
            code_source = getattr(vin, attr_name)

            code = code_source[attr_idx]

        defs = detail.defs

        if callable(defs):
            defs = defs(details)

        self._supported = bool(source)
        """Flag indicating that this detail extraction is available."""

        self.code: str = code
        self.name: Optional[str] = defs.get(code)

    def __str__(self):
        return self.name or self.code

    def __bool__(self):
        return self._supported


class Detail:
    """Vin detail descriptor."""

    __slots__ = ['source', 'defs']

    def __init__(self, code_source=None, definitions=None):
        self.source = code_source
        self.defs = definitions or {}

    def __get__(self, instance: 'VinDetails', owner: Type['VinDetails']) -> DetailWrapper:
        """
        :param instance:
        :param owner:

        """
        return DetailWrapper(instance, self)


class VinDetails(Annotatable):
    """Offers advanced (manufacturer specific) VIN data extraction ficilities."""

    annotate_titles = {
        'body': 'Body',
        'engine': 'Engine',
        'model': 'Model',
        'plant': 'Plant',
        'transmission': 'Transmission',
        'serial': 'Serial',
    }

    def __init__(self, vin: 'Vin'):
        self._vin = vin

    body = Detail()
    engine = Detail()
    model = Detail()
    plant = Detail()
    transmission = Detail()
    serial = Detail()



from typing import Dict


def __unpack_countries_map(counties: Dict[str, str]) -> Dict[str, str]:
    unpacked = {}

    seq = 'ABCDEFGHJKLMNPRSTUVWXYZ1234567890'

    for code, title in counties.items():
        first, _, span = code.partition('-')

        if span:
            ch_from, ch_till = span

        else:
            ch_from = 'A'
            ch_till = '0'

        for ch in seq[seq.index(ch_from):seq.index(ch_till) + 1]:
            unpacked[first + ch] = title

    return unpacked


# ISO 3780 https://www.iso.org/standard/45844.html
COUNTRIES = __unpack_countries_map({
    'A-AH': 'South Africa',
    'A-JK': 'Ivory Coast',
    'A-LM': 'Lesotho',
    'A-NP': 'Botswana',
    'A-RS': 'Namibia',
    'A-TU': 'Madagascar',
    'A-VW': 'Mauritius',
    'A-XY': 'Tunisia',
    'A-Z1': 'Cyprus',
    'A-23': 'Zimbabwe',
    'A-45': 'Mozambique',
    'B-AB': 'Angola',
    'B-FG': 'Kenya',
    'B-LL': 'Nigeria',
    'B-RR': 'Algeria',
    'B-34': 'Libya',
    'C-AB': 'Egypt',
    'C-FG': 'Morocco',
    'C-LM': 'Zambia',
    'H-': 'China',
    'J-': 'Japan',
    'K-FH': 'Israel',
    'K-LR': 'South Korea',
    'K-ST': 'Jordan',
    'L-': 'China',
    'M-AE': 'India',
    'M-FK': 'Indonesia',
    'M-LR': 'Thailand',
    'M-SS': 'Myanmar',
    'M-UU': 'Mongolia',
    'M-XX': 'Kazakhstan',
    'M-16': 'India',
    'N-AE': 'Iran',
    'N-FG': 'Pakistan',
    'N-JJ': 'Iraq',
    'N-LR': 'Turkey',
    'N-ST': 'Uzbekistan',
    'N-UU': 'Azerbaijan',
    'N-YY': 'Armenia',
    'N-15': 'Iran',
    'P-AC': 'Philippines',
    'P-FG': 'Singapore',
    'P-LR': 'Malaysia',
    'P-ST': 'Bangladesh',
    'R-AB': 'United Arab Emirates',
    'R-FK': 'Taiwan',
    'R-LN': 'Vietnam',
    'R-PP': 'Laos',
    'R-ST': 'Saudi Arabia',
    'R-UW': 'Russia',
    'R-11': 'Hong Kong',
    'S-AM': 'United Kingdom',
    'S-NT': 'Germany',
    'S-UZ': 'Poland',
    'S-12': 'Latvia',
    'S-33': 'South Ossetia',
    'T-AH': 'Switzerland',
    'T-JP': 'Czech Republic',
    'T-RV': 'Hungary',
    'T-W2': 'Portugal',
    'T-35': 'Serbia & Montenegro',
    'T-66': 'Andorra',
    'U-AC': 'Spain',
    'U-HM': 'Denmark',
    'U-NR': 'Ireland',
    'U-UW': 'Romania',
    'U-12': 'Macedonia',
    'U-57': 'Slovakia',
    'U-80': 'Bosnia & Herzogovina',
    'V-AE': 'Austria',
    'V-FR': 'France',
    'V-SW': 'Spain',
    'V-X2': 'Serbia',
    'V-35': 'Croatia',
    'V-68': 'Estonia',
    'W-': 'Germany',
    'X-AC': 'Bulgaria',
    'X-DE': 'Russia',
    'X-FH': 'Greece',
    'X-JK': 'Russia',
    'X-LR': 'Netherlands',
    'X-SW': 'Russia',
    'X-XY': 'Luxembourg',
    'X-Z0': 'Russia',
    'Y-AE': 'Belgium',
    'Y-FK': 'Finland',
    'Y-NN': 'Malta',
    'Y-SW': 'Sweden',
    'Y-X2': 'Norway',
    'Y-35': 'Belarus',
    'Y-68': 'Ukraine',
    'Z-AU': 'Italy',
    'Z-XZ': 'Slovenia',
    'Z-11': 'San Marino',
    'Z-35': 'Lithuania',
    'Z-60': 'Russia',
    '1-': 'United States',
    '2-A5': 'Canada',
    '3-AX': 'Mexico',
    '3-55': 'Dom. Republic',
    '3-66': 'Honduras',
    '3-77': 'Panama',
    '3-89': 'Puerto Rico',
    '4-': 'United States',
    '5-': 'United States',
    '6-AX': 'Australia',
    '6-Y1': 'New Zealand',
    '7-': 'United States',
    '8-AE': 'Argentina',
    '8-FG': 'Chile',
    '8-LN': 'Ecuador',
    '8-ST': 'Peru',
    '8-XZ': 'Venezuela',
    '8-22': 'Bolivia',
    '9-AE': 'Brazil',
    '9-FG': 'Colombia',
    '9-SV': 'Uruguay',
    '9-39': 'Brazil',
})



from typing import Dict, Any, Type

if False:  # pragma: nocover
    from .details._base import VinDetails  # noqa


class Annotatable:

    annotate_titles = {}

    def annotate(self) -> Dict[str, Any]:

        annotations = {}
        no_attr = set()

        for attr_name, label in self.annotate_titles.items():
            value = getattr(self, attr_name, no_attr)

            if value is no_attr:
                continue

            if isinstance(value, list):
                value = ', '.join(f'{val}' for val in value)

            annotations[label] = f'{value}'

        return dict((title, value) for title, value in sorted(annotations.items(), key=lambda item: item[0]))


class Brand:

    __slots__ = ['manufacturer']

    extractor: Type['VinDetails'] = None

    def __init__(self, manufacturer: str = None):
        self.manufacturer = manufacturer or self.title

    @property
    def title(self) -> str:
        return self.__class__.__name__

    def __str__(self):
        return f'{self.title} ({self.manufacturer})'


class UnsupportedBrand(Brand):
    """Unsupported brand."""



from .common import Brand
from .details import *


class Lada(Brand):

    extractor = AvtoVazDetails


class Nissan(Brand):

    extractor = NissanDetails


class Opel(Brand):

    extractor = OpelDetails


class Renault(Brand):

    extractor = RenaultDetails



from datetime import datetime
from itertools import cycle
import re
from typing import Optional, List

from .common import Annotatable, Brand, UnsupportedBrand
from .dicts import COUNTRIES, WMI, REGIONS
from .exceptions import ValidationError

if False:  # pragma: nocover
    from .details._base import VinDetails  # noqa


class Vin(Annotatable):
    """Offers basic VIN data extraction facilities."""

    annotate_titles = {
        'manufacturer': 'Manufacturer',
        'region': 'Region',
        'country': 'Country',
        'years': 'Years',
    }

    def __init__(self, num: str):
        self.num = self.validate(num)

        details_extractor = self.brand.extractor

        if details_extractor:
            details_extractor = details_extractor(self)

        self.details: VinDetails = details_extractor

    def __str__(self):
        return self.num

    @classmethod
    def validate(self, num: str) -> str:
        """Performs basic VIN validation and sanation.

        :param num:

        """
        num = num.strip().upper()

        num_len = len(num)
        if num_len != 17:
            raise ValidationError(f'VIN number requires 17 chars ({num_len} given)')

        pattern = r"^[A-HJ-NPR-Z0-9]{17}$"
        if not re.match(pattern, num):
            raise ValidationError(f"VIN number must only contain alphanumeric symbols except 'I', 'O', and 'Q' ")

        return num

    def verify_checksum(self, *, check_year: bool = True) -> bool:
        """Performs checksum verification.

        .. warning:: Not every manufacturer uses VIN checksum rules.

        :param check_year: Whether to also check the model year.
            Note that not all manufacturer abey the rule. Default: True.

        """
        if check_year and self.vis[0] in {'U', 'Z', '0'}:
            return False

        trans = {
            'A': 1, 'B': 2, 'C': 3, 'D': 4, 'E': 5, 'F': 6, 'G': 7, 'H': 8,
            'J': 1, 'K': 2, 'L': 3, 'M': 4, 'N': 5,         'P': 7,         'R': 9,
                    'S': 2, 'T': 3, 'U': 4, 'V': 5, 'W': 6, 'X': 7, 'Y': 8, 'Z': 9,
        }
        weights = (8, 7, 6, 5, 4, 3, 2, 10, 0, 9, 8, 7, 6, 5, 4, 3, 2)

        checksum = 0

        for pos, char in enumerate(self.num):
            value = int(trans.get(char, char))
            checksum += (value * weights[pos])

        checksum = int(checksum) % 11

        check_digit = 'X' if checksum == 10 else checksum

        return str(check_digit) == self.vds[5]

    @property
    def wmi(self) -> str:
        """WMI (World Manufacturers Identification)"""
        return self.num[:3]

    @property
    def brand(self) -> Brand:
        """Brand object."""

        wmi = self.wmi

        brand = WMI.get(wmi)

        if not brand:
            brand = WMI.get(wmi[:2])

        if isinstance(brand, str):
            brand = Brand(brand)

        if brand is None:
            brand = UnsupportedBrand()

        return brand

    @property
    def manufacturer(self) -> str:
        """Manufacturer title."""
        return self.brand.manufacturer

    @property
    def manufacturer_is_small(self) -> bool:
        """A manufacturer who builds fewer than 1000 vehicles per year."""
        return str(self.wmi[2]) == '9'

    @property
    def vds(self) -> str:
        """VDS (Vehicle Descriptor Section)"""
        return self.num[3:9]

    @property
    def vis(self) -> str:
        """VIS (Vehicle Identifier Section)"""
        return self.num[9:17]

    @property
    def region_code(self) -> str:
        return self.wmi[0]

    @property
    def region(self) -> Optional[str]:
        code = self.region_code

        title = None

        for chars, title_ in REGIONS.items():
            if code in chars:
                title = title_
                break

        return title

    @property
    def country_code(self) -> str:
        return self.wmi[0:2]

    @property
    def country(self) -> Optional[str]:
        return COUNTRIES.get(self.country_code)

    @property
    def years(self) -> List[int]:
        letters = 'ABCDEFGHJKLMNPRSTVWXY123456789'
        year_letter = self.vis[0]

        year = 1979
        year_current = datetime.now().year

        result = []

        for letter in cycle(letters):
            year += 1

            if letter == year_letter:
                result.append(year)

            if year == year_current:
                break

        result.sort(reverse=True)

        return result

    @property
    def squish_vin(self) -> str:
        """Squish (or Pattern) VIN.

        The first 11 digits of the VIN minus the 9th digit (positions 1-8,
        positions 10 and 11).

        Squish VIN encodes vehicle information while preventing its precise
        identification. It can be useful for anonymization and privacy purposes.
        """
        return self.num[:8] + self.num[9:11]



from collections import OrderedDict

from vininfo import Vin


def test_lada():

    vin = Vin('XTAGFK330JY144213')

    assert '%s' % vin
    assert vin.wmi == 'XTA'
    assert vin.manufacturer == 'AvtoVAZ'
    assert vin.vds == 'GFK330'
    assert vin.vis == 'JY144213'
    assert vin.years == [2018, 1988]
    assert vin.region_code == 'X'
    assert vin.region == 'Europe'
    assert vin.country_code == 'XT'
    assert vin.country == 'Russia'
    assert vin.annotate() == OrderedDict([
        ('Country', 'Russia'),
        ('Manufacturer', 'AvtoVAZ'),
        ('Region', 'Europe'),
        ('Years', '2018, 1988'),
    ])
    assert '%s' % vin.brand == 'Lada (AvtoVAZ)'

    details = vin.details
    assert details.model.code == 'F'
    assert details.model.name == 'Vesta'
    assert details.body.code == 'K'
    assert details.body.name == 'Station Wagon, 5-Door'
    assert details.engine.code == '3'
    assert details.engine.name == '21179'
    assert details.transmission.code == '3'
    assert details.transmission.name == 'Manual, 5-Gear (Renault JH3 514)'
    assert details.plant.code == 'Y'
    assert details.plant.name == 'Izhevsk'
    assert details.annotate() == OrderedDict([
        ('Body', 'Station Wagon, 5-Door'),
        ('Engine', '21179'),
        ('Model', 'Vesta'),
        ('Plant', 'Izhevsk'),
        ('Serial', '144213'),
        ('Transmission', 'Manual, 5-Gear (Renault JH3 514)'),
    ])
    assert not vin.verify_checksum()



from ..brands import Lada, Nissan, Opel, Renault

# NOTE:
# if you want to extend this mapping with new WMIs, please use
# vininfo.utils.merge_wmi() function and replace the entire WMI with
# the result of this function. This comment must stay intact.
WMI = {
    '000': 'Maserati',
    '04W': 'Buick',
    '0VF': 'Ford',
    '112': 'Volkswagen',
    '115': 'Mercedes-Benz',
    '117': 'Volkswagen',
    '119': 'Replica/Kit Makes',
    '123': 'Mercedes-Benz',
    '124': 'Chevrolet',
    '137': 'AM',
    '178': 'Jaguar',
    '17V': 'Ford',
    '19': 'Acura',
    '19X': 'Honda',
    '1A4': 'Chrysler',
    '1A8': 'Chrysler',
    '1B': 'Dodge',
    '1C': 'Chrysler',
    '1C4': 'Dodge',
    '1C6': 'Ram',
    '1C8': 'Chrysler',
    '1CN': 'Lincoln',
    '1D': 'Dodge',
    '1F': 'Ford',
    '1F2': 'Subaru',
    '1F9': 'FWD Corp.',
    '1FU': 'Freightliner',
    '1FV': 'Freightliner',
    '1G': 'General Motors',
    '1G1': 'Chevrolet',
    '1G2': 'Pontiac',
    '1G3': 'Oldsmobile',
    '1G4': 'Buick',
    '1G6': 'Cadillac',
    '1G8': 'Saturn',
    '1G9': 'Google',
    '1GA': 'Chevrolet',
    '1GB': 'Chevrolet USA',
    '1GC': 'Chevrolet',
    '1GD': 'GMC',
    '1GE': 'Cadillac',
    '1GG': 'Isuzu',
    '1GH': 'Oldsmobile',
    '1GJ': 'GMC',
    '1GK': 'GMC',
    '1GM': 'Pontiac',
    '1GN': 'Chevrolet USA',
    '1GT': 'GMC Truck',
    '1GY': 'Cadillac',
    '1H': 'Honda',
    '1HD': 'Harley-Davidson',
    '1J': 'Jeep',
    '1J4': 'Jeep',
    '1J8': 'Jeep',
    '1JN': Nissan(),
    '1L': 'Lincoln',
    '1L1': 'Lincoln',
    '1L4': 'Jeep',
    '1LN': 'Lincoln',
    '1M': 'Mercury',
    '1M1': 'Mack Truck',
    '1M2': 'Mack Truck',
    '1M3': 'Mack Truck',
    '1M4': 'Mack Truck',
    '1M9': 'Mynatt Truck & Equipment',
    '1ME': 'Mercury',
    '1N': Nissan(),
    '1NX': 'NUMMI',
    '1P3': 'Plymouth',
    '1P4': 'Plymouth',
    '1R9': 'Roadrunner Hay Squeeze',
    '1TK': 'Scion',
    '1V1': 'Volkswagen USA (Commercials)',
    '1V2': 'Volkswagen',
    '1V4': Nissan(),
    '1VW': 'Volkswagen',
    '1XK': 'Kenworth',
    '1XP': 'Peterbilt',
    '1Y1': 'Chevrolet',
    '1YV': 'Mazda',
    '1Z3': 'Mitsubishi',
    '1Z7': 'Mitsubishi',
    '1ZV': 'Auto Alliance International',
    '1ZW': 'Mercury',
    '210': 'Ford',
    '2A4': 'Chrysler Canada',
    '2A8': 'Chrysler Canada',
    '2B3': 'Dodge Canada',
    '2B4': 'Dodge',
    '2B5': 'Dodge',
    '2B6': 'Dodge',
    '2B7': 'Dodge',
    '2B8': 'Dodge',
    '2C3': 'Chrysler',
    '2C4': 'Chrysler Canada',
    '2C7': 'Dodge',
    '2C8': 'Chrysler',
    '2CK': 'Pontiac',
    '2CN': 'CAMI',
    '2CT': 'General Motors',
    '2D3': 'Dodge',
    '2D4': 'Dodge Canada',
    '2D6': 'Dodge',
    '2D7': 'Dodge',
    '2D8': 'Dodge Canada',
    '2DG': 'Ontario Drive & Gear',
    '2F': 'Ford',
    '2FU': 'Freightliner',
    '2FV': 'Freightliner',
    '2FZ': 'Sterling',
    '2G': 'General Motors',
    '2G1': 'Chevrolet',
    '2G2': 'Pontiac',
    '2G3': 'Oldsmobile',
    '2G4': 'Buick',
    '2G6': 'Cadillac',
    '2G9': 'Gnome Homes',
    '2GC': 'Chevrolet Canada',
    '2GE': 'Cadillac',
    '2GK': 'GMC',
    '2GN': 'Chevrolet Canada',
    '2GT': 'GMC',
    '2H': 'Honda',
    '2HH': 'Acura',
    '2HM': 'Hyundai',
    '2HN': 'Acura',
    '2L': 'Lincoln',
    '2L1': 'Lincoln',
    '2LM': 'Lincoln',
    '2LN': 'Lincoln',
    '2M': 'Mercury',
    '2NV': 'Nova Bus',
    '2P3': 'Plymouth',
    '2P4': 'Plymouth',
    '2S2': 'Suzuki',
    '2S3': 'Suzuki Canada',
    '2T': 'Toyota',
    '2T2': 'Lexus Canada',
    '2V4': 'Volkswagen',
    '2V8': 'Volkswagen',
    '2W': 'Western Star',
    '309': 'Chevrolet',
    '3A': 'Chrysler Mexico',
    '3A4': 'Chrysler',
    '3A8': 'Chrysler',
    '3B6': 'Dodge',
    '3B7': 'Dodge Mexico',
    '3C': 'Chrysler',
    '3C3': 'Fiat',
    '3C4': 'Dodge Mexico',
    '3C6': 'Ram',
    '3C7': 'Ram',
    '3C8': 'Chrysler',
    '3CZ': 'Honda Mexico',
    '3D': 'Dodge',
    '3F': 'Ford',
    '3G': 'General Motors',
    '3G0': 'Saab',
    '3G1': 'Chevrolet',
    '3G2': 'Pontiac',
    '3G5': 'Buick',
    '3G7': 'Pontiac',
    '3GC': 'Chevrolet Mexico',
    '3GD': 'GMC',
    '3GK': 'GMC',
    '3GN': 'Chevrolet Mexico',
    '3GS': 'Saturn',
    '3GT': 'GMC',
    '3GV': 'Chevrolet',
    '3GY': 'Cadillac',
    '3H': 'Honda',
    '3KP': 'Kia',
    '3LN': 'Lincoln',
    '3MD': 'Mazda',
    '3ME': 'Mercury Mexico',
    '3MY': 'Mazda Mexico',
    '3MZ': 'Mazda Mexico',
    '3N': Nissan(),
    '3N6': 'Chevrolet',
    '3P3': 'Plymouth Mexico',
    '3TM': 'Toyota Mexico',
    '3VV': 'Volkswagen',
    '3VW': 'Volkswagen',
    '460': 'Mercedes-Benz',
    '4A': 'Mitsubishi',
    '4B3': 'Dodge',
    '4C3': 'Chrysler',
    '4F': 'Mazda',
    '4G1': 'Chevrolet',
    '4GD': Opel(),
    '4J': 'Mercedes-Benz',
    '4M': 'Mercury',
    '4N2': Nissan(),
    '4NU': 'Isuzu',
    '4RK': 'Nova Bus',
    '4S': 'Subaru-Isuzu Automotive',
    '4S2': 'Isuzu',
    '4S3': 'Subaru',
    '4S4': 'Subaru',
    '4S6': 'Honda',
    '4T': 'Toyota',
    '4US': 'BMW',
    '4UZ': 'Frt-Thomas Bus',
    '4V': 'Volvo',
    '54D': 'Chevrolet',  # Spartan,
    '55': 'Mercedes-Benz',
    '55S': 'Mercedes-Benz',
    '58A': 'Lexus',
    '5BZ': Nissan(),
    '5F': 'Honda',
    '5FR': 'Acura',
    '5GA': 'Buick',
    '5GN': 'Hummer',
    '5GR': 'Hummer',
    '5GT': 'Hummer',
    '5GZ': 'Saturn',
    '5J6': 'Honda',
    '5J8': 'Acura',
    '5KB': 'Honda',
    '5L': 'Lincoln',
    '5N1': Nissan(),
    '5N3': Nissan('Infiniti'),
    '5NA': Nissan(),
    '5NM': 'Hyundai',
    '5NP': 'Hyundai',
    '5S3': 'Saab',
    '5T': 'Toyota',
    '5U': 'BMW',
    '5X': 'Hyundai/Kia',
    '5XX': 'Kia',
    '5XY': 'Kia',
    '5Y2': 'Pontiac',
    '5YF': 'Toyota',
    '5YJ': 'Tesla',
    '5YM': 'BMW',
    '5Z6': 'Suzuki',
    '601': 'Replica/Kit Makes',
    '602': 'Toyota',
    '6AB': 'MAN',
    '6F': 'Ford',
    '6F4': Nissan('Nissan Motor Company'),
    '6F5': 'Kenworth',
    '6FP': 'Ford Motor Company',
    '6G': 'General Motors',
    '6G1': 'Chevrolet',
    '6G2': 'Pontiac',
    '6G3': 'Chevrolet Australia',
    '6H': 'Holden',
    '6H8': 'General Motors-Holden',
    '6MM': 'Mitsubishi',
    '6T1': 'Toyota',
    '7A3': 'Honda',
    '7FA': 'Honda',
    '8A1': Renault(),
    '8AC': 'Mercedes Benz',
    '8AD': 'Peugeot',
    '8AF': 'Ford',
    '8AG': 'General Motors',
    '8AJ': 'Toyota',
    '8AK': 'Suzuki',
    '8AP': 'Fiat',
    '8AT': 'Iveco',
    '8AW': 'Volkswagen',
    '8BC': 'Citroën',
    '8BR': 'Mercedes-Benz Argentina',
    '8BT': 'Mercedes-Benz Argentina',
    '8C3': 'Honda',
    '8GD': 'Peugeot',
    '8GG': 'Chevrolet',
    '935': 'Citroën',
    '936': 'Peugeot',
    '93H': 'Honda',
    '93R': 'Toyota',
    '93U': 'Audi',
    '93V': 'Audi Brazil',
    '93W': 'Fiat Professional',
    '93X': 'Souza Ramos - Mitsubishi / Suzuki',
    '93Y': Renault(),
    '93Z': 'Iveco',
    '94D': Nissan(),
    '953': 'VW Trucks / MAN',
    '95P': 'CAOA / Hyundai',
    '988': 'Jeep',
    '98M': 'BMW',
    '98R': 'Chery',
    '99A': 'Audi',
    '99J': 'JLR Jaguar Land Rover',
    '9BD': 'Fiat Automóveis',
    '9BF': 'Ford',
    '9BG': 'General Motors',
    '9BH': 'Hyundai Motor Company / Hyundai',
    '9BM': 'Mercedes Benz',
    '9BR': 'Toyota',
    '9BS': 'Scania',
    '9BV': 'Volvo Trucks',
    '9BW': 'Volkswagen',
    '9C2': 'Honda Motorcycles',
    '9C6': 'Yamaha',
    '9CD': 'Suzuki Motorcycles',
    '9FB': Renault(),
    '9UJ': 'Chery',
    '9UK': 'Lifan',
    '9UW': 'Kia',
    'AAV': 'Volkswagen',
    'AFA': 'Ford',
    'AHT': 'Toyota',
    'B01': 'Cadillac',
    'CF1': Renault(),
    'CL9': 'Wallyscar',
    'DA4': 'Jeep',
    'EDB': 'Mercedes-Benz',
    'FM2': 'Mercury',
    'FSM': 'FSM',
    'FV1': Renault(),
    'FV3': 'Peugeot',
    'FV7': 'Citroen',
    'G4G': 'Buick',
    'G61': 'Cadillac',
    'GA1': Renault(),
    'GKA': 'GMC',
    'JA': 'Isuzu',
    'JA3': 'Mitsubishi',
    'JA4': 'Mitsubishi',
    'JAC': 'Isuzu',
    'JAE': 'Acura',
    'JB3': 'Dodge',
    'JC1': 'Fiat Automobiles/Mazda',
    'JDA': 'Daihatsu',
    'JF': 'Fuji Heavy Industries',
    'JF1': 'Scion',
    'JF2': 'Subaru',
    'JF4': 'Saab',
    'JFS': 'Subaru',
    'JGN': 'Chevrolet',
    'JH': 'Honda',
    'JH4': 'Acura',
    'JHM': 'Honda',
    'JK': 'Kawasaki',
    'JM': 'Mazda',
    'JMB': 'Mitsubishi',
    'JN': Nissan(),
    'JNK': Nissan('Infiniti'),
    'JNR': Nissan('Infiniti'),
    'JNT': Nissan('Infiniti'),
    'JNX': Nissan('Infiniti'),
    'JS': 'Suzuki',
    'JT': 'Toyota',
    'JT6': 'Lexus',
    'JT8': 'Lexus',
    'JTD': 'Toyota',
    'JTE': 'Toyota',
    'JTH': 'Lexus',
    'JTJ': 'Lexus',
    'JTK': 'Scion',
    'JTL': 'Scion',
    'JTM': 'Toyota',
    'JTN': 'Scion',
    'JTS': 'Toyota',
    'JY': 'Yamaha',
    'KC4': 'Dodge',
    'KL': 'Daewoo/GM Korea',
    'KL1': 'Chevrolet',
    'KL2': 'Pontiac',
    'KL4': 'Buick',
    'KL5': 'Suzuki',
    'KL7': 'Chevrolet',
    'KL8': 'Chevrolet',
    'KLA': 'Chevrolet',
    'KLM': 'Lincoln',
    'KM': 'Hyundai',
    'KMH': 'Genesis',
    'KN': 'Kia',
    'KN1': Nissan(),
    'KNM': Renault('Renault Samsung'),
    'KP': 'SsangYong',
    'KRX': 'BMW',
    'L4C': 'Buick',
    'L56': Renault('Renault Samsung'),
    'L5Y': 'Merato Motorcycle Taizhou Zhongneng',
    'L6T': 'Geely',
    'LBE': 'Beijing Hyundai',
    'LBV': 'BMW Brilliance',
    'LC0': 'BYD Bus',
    'LDC': 'Dongfeng Peugeot-Citroën',
    'LDY': 'Zhongtong Coach',
    'LE4': 'Beijing Benz',
    'LFM': 'FAW Toyota',
    'LFP': 'FAW Car',
    'LFV': 'FAW-Volkswagen',
    'LGB': 'Dongfeng Nissan',
    'LGH': 'Dong Feng (DFM), China',
    'LGJ': 'Dongfeng Fengshen',
    'LGW': 'Great Wall (Havel)',
    'LGX': 'BYD Auto',
    'LH1': 'FAW Haima',
    'LHG': 'Guangzhou Honda',
    'LJ1': 'JAC',
    'LJD': 'Dongfeng Yueda Kia',
    'LKL': 'Suzhou King Long',
    'LLV': 'Lifan',
    'LMG': 'GAC Trumpchi',
    'LPA': 'Changan PSA (DS Automobiles)',
    'LRB': 'Buick China',
    'LS5': 'Changan Suzuki',
    'LSG': 'SAIC General Motors',
    'LSJ': 'SAIC MG',
    'LSV': 'SAIC Volkswagen',
    'LSY': 'Brilliance Zhonghua',
    'LTV': 'FAW Toyota (Tianjin)',
    'LUC': 'Honda',
    'LVG': 'GAC Toyota',
    'LVH': 'Dongfeng Honda',
    'LVR': 'Changan Mazda',
    'LVS': 'Changan Ford',
    'LVV': 'Chery',
    'LVY': 'Volvo',
    'LWV': 'GAC Fiat',
    'LYV': 'Volvo China',
    'LZE': 'Isuzu Guangzhou',
    'LZG': 'Shaanxi Automobile Group',
    'LZM': 'MAN',
    'LZW': 'SAIC GM Wuling',
    'LZY': 'Yutong',
    'MA1': 'Mahindra',
    'MA3': 'Suzuki',
    'MA7': 'Honda Siel Cars',
    'MAJ': 'FordS',
    'MAL': 'Hyundai',
    'MAT': 'Tata',
    'MBH': Nissan(),
    'MC2': 'Volvo Eicher commercial vehicles limited.',
    'MDH': Nissan(),
    'MHR': 'Honda',
    'ML3': 'Mitsubishi Thailand',
    'MM0': 'Mazda',
    'MM8': 'Mazda',
    'MMB': 'Mitsubishi',
    'MMC': 'Mitsubishi',
    'MMM': 'Chevrolet',
    'MMS': 'Suzuki',
    'MMT': 'Mitsubishi',
    'MNB': 'Ford',
    'MNT': Nissan(),
    'MP1': 'Isuzu',
    'MPA': 'Isuzu',
    'MR0': 'Toyota',
    'MRH': 'Honda',
    'MS0': 'KIA Myanmar',
    'NLA': 'Honda',
    'NLE': 'Mercedes-Benz Turk Truck',
    'NLH': 'Hyundai',
    'NLJ': 'Hyundai',
    'NM0': 'Ford Otosan',
    'NM4': 'Tofas Turk',
    'NMT': 'Toyota',
    'PE1': 'Ford',
    'PE3': 'Mazda',
    'PL1': 'Proton',
    'SAD': 'Jaguar',
    'SAH': 'Honda',
    'SAJ': 'Jaguar',
    'SAL': 'Land Rover',
    'SAR': 'Rover',
    'SAT': 'Triumph',
    'SAX': 'Rover',
    'SB1': 'Toyota',
    'SBM': 'Mclaren',
    'SCA': 'Rolls Royce',
    'SCB': 'Bentley',
    'SCC': 'Lotus Cars',
    'SCE': 'DeLorean',
    'SCF': 'Aston Martin Lagonda Limited',
    'SDB': 'Peugeot UK',
    'SED': Opel(),
    'SEY': 'LDV',
    'SFA': 'Ford',
    'SFD': 'Alexander Dennis',
    'SHH': 'Honda',
    'SHS': 'Honda',
    'SJA': 'Bentley',
    'SJK': Nissan('Infiniti'),
    'SJN': Nissan(),
    'SKF': Opel(),
    'SMT': 'Triumph',
    'SNE': 'Jeep',
    'SNT': 'Honda',
    'STE': 'Toyota',
    'SU9': 'Solaris Bus & Coach',
    'SUF': 'Fiat Auto Poland / FSM',
    'SUL': 'Daewoo Poland / FSO',
    'SUP': 'Daewoo Poland / FSO',
    'SUR': 'Land Rover',
    'TC3': 'Chrysler',
    'TCC': 'Micro Compact Car AG (SMART 1998-1999)',
    'TDM': 'QUANTYA Swiss Electric Movement',
    'TK9': 'SOR',
    'TM9': 'Škoda trolleybuses',
    'TMA': 'Hyundai',
    'TMB': 'Škoda',
    'TMK': 'Karosa',
    'TMP': 'Škoda trolleybuses',
    'TMT': 'Tatra',
    'TN9': 'Karosa',
    'TNB': 'Skoda',
    'TRA': 'Ikarus Bus',
    'TRU': 'Audi',
    'TSE': 'Ikarus Egyedi Autobuszgyar',
    'TSM': 'Suzuki',
    'TYB': 'Mitsubishi',
    'U5Y': 'Kia',
    'U6Y': 'Kia',
    'USY': 'Kia',
    'UU': 'Dacia',
    'UU1': Renault('Renault Dacia'),
    'V0L': Opel(),
    'VA0': 'ÖAF',
    'VBK': 'KTM',
    'VF0': 'Ford',
    'VF1': Renault(),
    'VF2': Renault(),
    'VF3': 'Peugeot',
    'VF4': 'Talbot',
    'VF5': 'Iveco Unic SA',
    'VF6': 'Renault Trucks/Volvo',
    'VF7': 'Citroën',
    'VF8': 'Matra/Talbot/Simca',
    'VF9': 'Bugatti',
    'VFB': Renault(),
    'VFE': 'IvecoBus',
    'VFF': 'Peugeot',
    'VFG': 'Citroen',
    'VFJ': Renault(),
    'VFZ': 'Citroen',
    'VH8': 'Microcar',
    'VLG': 'Aixam',
    'VLU': 'Scania',
    'VN1': Opel(),
    'VNE': 'Irisbus',
    'VNK': 'Toyota',
    'VNV': Renault(),
    'VR7': 'Citroën',
    'VS1': 'Iveco',
    'VS3': 'Peugeot',
    'VS5': Renault(),
    'VS6': 'Ford',
    'VS7': 'Citroen',
    'VS9': 'Carrocerias Ayats',
    'VSA': 'Mercedes-Benz',
    'VSE': 'Suzuki / Santana Motors',
    'VSK': Nissan(),
    'VSS': 'SEAT',
    'VSX': Opel(),
    'VSY': Renault(),
    'VSZ': 'Seat',
    'VV9': 'Tauro Sport Auto',
    'VW1': Renault(),
    'VW2': 'Volkswagen',
    'VWA': Nissan(),
    'VWG': 'Volkswagen Spain',
    'VWV': 'Volkswagen',
    'VX1': 'Zastava / Yugo',
    'VY1': 'Volvo',
    'W00': Opel(),
    'W04': 'Buick',
    'W06': 'Cadillac',
    'W08': 'Saturn',
    'W09': 'Ruf Automobile',
    'W0L': Opel('Opel/Vauxhall'),
    'W0S': Opel('Opel Special Vehicles'),
    'W0V': Opel(),
    'W1': 'Mercedes-Benz',
    'WA1': 'Audi',
    'WAG': 'Neoplan',
    'WAP': 'Alpina',
    'WAU': 'Audi',
    'WAV': 'Audi',
    'WAX': 'SsangYong',
    'WB': 'BMW',
    'WBA': 'BMW',
    'WBD': 'Mercedes-Benz',
    'WBS': 'BMW M',
    'WBX': 'BMW',
    'WBY': 'BMW',
    'WCD': 'Mercedes-Benz (Sprinter)',
    'WD0': 'Dodge',
    'WD2': 'Dodge',
    'WD3': 'Daimler AG (Sprinter)',
    'WD4': 'Daimler AG (Sprinter)',
    'WD5': 'Dodge',
    'WD8': 'Mercedes-Benz',
    'WDA': 'Daimler AG (Sprinter)',
    'WDB': 'Mercedes-Benz',
    'WDC': 'DaimlerChrysler AG/Daimler AG',
    'WDD': 'DaimlerChrysler AG/Daimler AG',
    'WDF': 'Mercedes-Benz',
    'WDP': 'Mercedes-Benz (Sprinter)',
    'WDR': 'Mercedes-Benz (Sprinter)',
    'WDW': 'Dodge',
    'WDX': 'Dodge',
    'WDY': 'Mercedes-Benz (Sprinter)',
    'WDZ': 'Mercedes-Benz (Sprinter)',
    'WE0': 'Ford',
    'WEB': 'EvoBus',
    'WF0': 'Ford of Europe',
    'WF1': Renault(),
    'WF3': 'Peugeot',
    'WF7': 'Citroen',
    'WFD': 'Fliegl',
    'WFO': 'Ford',
    'WJM': 'Iveco',
    'WJR': 'Irmscher',
    'WKK': 'Karl Kässbohrer Fahrzeugwerke',
    'WMA': 'MAN',
    'WMB': 'Audi',
    'WME': 'Smart',
    'WMW': 'Mini',
    'WMX': 'DaimlerChrysler AG/Daimler AG',
    'WMZ': 'MINI',
    'WNK': 'Toyota',
    'WOL': Opel(),
    'WP0': 'Porsche car',
    'WP1': 'Porsche SUV',
    'WS0': 'Ford',
    'WSS': 'Seat',
    'WUA': 'Quattro',
    'WUW': 'Volkswagen',
    'WV': 'Volkswagen',
    'WV0': 'Ford',
    'WV1': 'Volkswagen Commercial Vehicles',
    'WV2': 'Volkswagen Commercial Vehicles',
    'WV3': 'Volkswagen Trucks',
    'WVG': 'Volkswagen',
    'WVW': 'Volkswagen',
    'WVZ': 'Volkswagen',
    'WWD': 'Mercedes-Benz',
    'WWW': 'Volkswagen',
    'WYG': 'Volkswagen',
    'WYW': 'Volkswagen',
    'WZW': 'Volkswagen',
    'X7L': Renault(),
    'X96': 'Mercedes-Benz',
    'XL9': 'Spyker',
    'XLB': 'Volvo',
    'XLR': 'DAF Trucks',
    'XMC': 'Mitsubishi (NedCar)',
    'XNC': 'Mitsubishi',
    'XTA': Lada('AvtoVAZ'),
    'XUF': Opel(),
    'XW8': 'Volkswagen',
    'XWE': 'Kia',
    'XWF': Opel(),
    'XXX': 'Honda',
    'Y6D': Opel(),
    'YAR': 'Toyota',
    'YCM': 'Mazda',
    'YH4': 'Fisker',
    'YK1': 'Saab',
    'YMB': 'Skoda',
    'YS2': 'Scania, Södertälje',
    'YS3': 'Saab',
    'YS4': 'Scania, Katrineholm',
    'YTN': 'Saab NEVS',
    'YV1': 'Volvo Cars',
    'YV2': 'Volvo Trucks',
    'YV3': 'Volvo Buses',
    'YV4': 'Volvo Cars',
    'Z12': Opel(),
    'Z3B': 'Chevrolet',
    'ZA9': 'Bugatti',
    'ZAC': 'FCA',
    'ZAF': 'Fiat',
    'ZAM': 'Maserati',
    'ZAP': 'Piaggio/Vespa/Gilera',
    'ZAR': 'Alfa Romeo',
    'ZAS': 'Alfa',
    'ZCF': 'Iveco',
    'ZCG': 'Cagiva SpA',
    'ZD4': 'Aprilia',
    'ZDF': 'Ferrari Dino',
    'ZDM': 'Ducati Motor Holdings SpA',
    'ZFA': 'Fiat Automobiles',
    'ZFB': 'Fiat',
    'ZFC': 'Fiat V.I.',
    'ZFF': 'Ferrari',
    'ZGA': 'IvecoBus',
    'ZHW': 'Lamborghini',
    'ZLA': 'Lancia',
    'ZN6': 'Maserati',
    'ZOM': 'OM',
    'ZSA': 'Fiat',
}



from vininfo import Vin


def test_renault():

    vin = Vin('VF14SRAP451234567')

    assert '%s' % vin
    assert vin.wmi == 'VF1'
    assert vin.manufacturer == 'Renault'
    assert vin.vds == '4SRAP4'
    assert vin.vis == '51234567'
    assert vin.years == [2005]
    assert vin.region_code == 'V'
    assert vin.region == 'Europe'
    assert vin.country_code == 'VF'
    assert vin.country == 'France'
    assert '%s' % vin.brand == 'Renault (Renault)'

    details = vin.details
    assert not details.engine
    assert details.model
    assert details.model.code == 'S'
    assert details.model.name == ['Logan', 'Sandero', 'Duster', 'Dokker', 'Lodgy']
    assert details.body.code == '4'
    assert details.body.name == 'Sedan, 4-Door'
    assert details.plant.code == 'P'
    assert details.plant.name == 'Mexico'
    assert details.transmission.code == '4'
    assert details.transmission.name == 'Manual, 5-Gears'
    assert details.serial.code == '1234567'


def test_bogus():
    vin = Vin('VF1KG1PBE34488860')
    details = vin.details
    assert details.engine.code == ''
    assert not details.engine
